// Released under the MIT licence.
// See LICENCE.txt for details.

#include "Backend.h"
#include "../../Misc.h"

#include <csignal>
#include <stdint.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <stdbool.h>
#include <sys/audioio.h>
#include <pthread.h>

static const int bufferSize = 300;
static const int samplingRate = 8000;

static pthread_mutex_t mixerMutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t organyaMutex = PTHREAD_MUTEX_INITIALIZER;

static int buffersWritten = 0;
static int sndfp;
static FILE* sndfile;
static void (*parent_callback)(long *stream, size_t frames_total);

static inline void setUpSoundFrame(long *mix) {
    memset(mix, 0, bufferSize * 2 * sizeof(long));
    parent_callback(mix, bufferSize);
}

static inline int soundWait() {
    // Wait until at least one audio buffer has been played back.
    // Return the number of audio buffers we have ever finished playing.

    audio_info_t audioInfo;
    ioctl(sndfp, AUDIO_GETINFO, &audioInfo);

    // audioInfo.play.eof is the number of "end-of-file records" the sound chip
    // has encountered. A "end-of-file record" is created by performing a
    // write() of zero length. We end each audio buffer with one such write.
    int previousEOF = audioInfo.play.eof;

    while(audioInfo.play.eof - previousEOF < 1) {
        if(previousEOF >= buffersWritten) {
            Backend_PrintError("Audio underrun.\n");
            break;
        }

        Backend_Delay(10); // Wait some reasonable time

        ioctl(sndfp, AUDIO_GETINFO, &audioInfo); // Get the new current EOF
    }
    return previousEOF;
}

static inline void feedOutSound(long* mix) {
    for(int i = 0; i < bufferSize * 2; i++) {
        short sample = mix[i];
        if(mix[i] > 0x7FFF) {
            sample = 0x7FFF;
        }
        if(mix[i] < -0x7FFF) {
            sample = -0x7FFF;
        }
        fputc((sample >> 8) & 0xFF, sndfile);
        fputc(sample & 0xFF, sndfile);
    }

    // Write a "end-of-file record" here used as a end-of-buffer marker. The
    // sound chip counts how many of these it has encountered and we can use
    // that to tell how many sound buffers it's played.
    write(sndfp, 0, 0);
    fflush(sndfile);

    buffersWritten++;
}

static void* soundThread(void*) {
    long mix[bufferSize * 2];
    Backend_PrintInfo("Entered sound thread.\n");

    while(1) {
        // Waiting for more to be needed
        int eofsPlayed = soundWait();

        // Produce audio until we're three buffers ahead of playback
        while(buffersWritten - eofsPlayed < 3) {
            setUpSoundFrame(mix);
            feedOutSound(mix);
        }
    }
}

unsigned long SoftwareMixerBackend_Init(void (*callback)(long *stream, size_t frames_total)) {
    parent_callback = callback;

    sndfile = fopen("/dev/audio", "wb");

    if(sndfile == NULL) {
        Backend_PrintError("Failed to open audio device.\n");
        return 0;
    }

    sndfp = fileno(sndfile);
    audio_info_t audioInfo;
    ioctl(sndfp, AUDIO_GETINFO, &audioInfo);
    audio_prinfo_t info = audioInfo.play;

    info.sample_rate = samplingRate;
    info.channels = 2;
    info.precision = 16;
    info.encoding = AUDIO_ENCODING_LINEAR;
    info.port = AUDIO_HEADPHONE;
    info.gain = 128;

    audioInfo.play = info;
    ioctl(sndfp, AUDIO_SETINFO, &audioInfo);

    pthread_mutex_init(&mixerMutex, NULL);
    pthread_mutex_init(&organyaMutex, NULL);

    pthread_t thread;
    pthread_attr_t thread_attr;
    pthread_attr_init(&thread_attr);

    pthread_create(&thread, &thread_attr, soundThread, (void*)NULL);

    return samplingRate; // Return the sampling rate
}

void SoftwareMixerBackend_Deinit(void) {
    close(sndfp);
}

bool SoftwareMixerBackend_Start(void) {
    return true;
}

void SoftwareMixerBackend_LockMixerMutex(void) {
    pthread_mutex_lock(&mixerMutex);
}

void SoftwareMixerBackend_UnlockMixerMutex(void) {
    pthread_mutex_unlock(&mixerMutex);
}

void SoftwareMixerBackend_LockOrganyaMutex(void) {
    pthread_mutex_lock(&organyaMutex);
}

void SoftwareMixerBackend_UnlockOrganyaMutex(void) {
    pthread_mutex_unlock(&organyaMutex);
}
