#include "../Software.h"
#include "../../../Misc.h"

#include <X11/X.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>

static Window window;

// Framebuffer is the image that the game gets to write to
static XImage* framebufferImage;
static char *framebufferPixels;
static XShmSegmentInfo framebufferSegInfo;
// Display is the one X gets to read from, we swap them every frame
static XImage* displayImage;
static char *displayPixels;
static XShmSegmentInfo displaySegInfo;

static int framebufferPitch;
static char* xfriendlyFB;
static GC gc;
static int screenWidth;
static int screenHeight;
static bool useShmP = false;

extern Display* xDisplay;
extern Visual* xVisual;

static bool CheckShm() {
    bool exists = XShmQueryExtension(xDisplay);
    return exists;
}

bool checkImageOK(XImage* image) {
    bool ok = true;
    ok &= (image->width == screenWidth);
    ok &= (image->bytes_per_line == screenWidth * 4);
    ok &= (image->depth == 24 || image->depth == 32);
    return ok;
}

void createFramebuffer() {
    gc = DefaultGC(xDisplay, DefaultScreen(xDisplay));

    int depth = DefaultDepth(xDisplay, DefaultScreen(xDisplay));

    // Setting up the framebuffer
    displayPixels = NULL;

    if(useShmP) {
        displayImage = XShmCreateImage(xDisplay, xVisual,
                                           depth, ZPixmap, NULL,
                                           &displaySegInfo, screenWidth, screenHeight);
        displaySegInfo.shmid = shmget(IPC_PRIVATE,
                                      displayImage->bytes_per_line * displayImage->height,
                                      IPC_CREAT | 0777);
        displaySegInfo.shmaddr = displayPixels = displayImage->data =
            (char*)shmat(displaySegInfo.shmid, NULL, 0);
        XShmAttach(xDisplay, &displaySegInfo);

        framebufferImage = XShmCreateImage(xDisplay, xVisual,
                                           depth, ZPixmap, NULL,
                                           &framebufferSegInfo, screenWidth, screenHeight);
        framebufferSegInfo.shmid = shmget(IPC_PRIVATE,
                                          framebufferImage->bytes_per_line * displayImage->height,
                                          IPC_CREAT | 0777);
        framebufferSegInfo.shmaddr = framebufferPixels = framebufferImage->data =
            (char*)shmat(framebufferSegInfo.shmid, NULL, 0);
        XShmAttach(xDisplay, &framebufferSegInfo);
    } else {
        displayPixels = (char*)malloc(screenWidth * screenHeight * 4);
        framebufferPixels = displayPixels;
        displayImage = XCreateImage(xDisplay, xVisual,
                                        depth, ZPixmap, 0,
                                        (char*)displayPixels, screenWidth, screenHeight,
                                        32, 0);
    }
}

bool WindowBackend_Software_CreateWindow(const char *window_title, size_t screen_width, size_t screen_height, bool fullscreen) {
    screenWidth = screen_width;
    screenHeight = screen_height;

    int screen = DefaultScreen(xDisplay);
    Window rootWindow = RootWindow(xDisplay, screen);

    useShmP = CheckShm();
    if(useShmP) {
        Backend_PrintInfo("Using X shared memory extension for higher image transfer speeds.\n");
    } else {
        Backend_PrintInfo("Unable to use X shared memory extension. Expect lower performance.\n");
        Backend_PrintInfo("On Sun, you might need to do the following:\n\nsetenv DISPLAY :0\nsetenv XSUNTRANSPORT shmem\nsetenv XSUNSMESIZE 512\n");
    }

    window = XCreateWindow(xDisplay, rootWindow, 0, 0, screen_width, screen_height, 0, 24, InputOutput, xVisual, 0, NULL);
    XSelectInput(xDisplay, window, ExposureMask | KeyPressMask | KeyReleaseMask);
    XMapRaised(xDisplay, window);
    XStoreName(xDisplay, window, window_title);

    // Make the close button work. This is also required for the window to show up.
    // weird
    Atom WM_DELETE_WINDOW = XInternAtom(xDisplay, "WM_DELETE_WINDOW", false);
    XSetWMProtocols(xDisplay, window, &WM_DELETE_WINDOW, 1);

    createFramebuffer();
    if(checkImageOK(displayImage) == false) {
        Backend_PrintError("Something's off with the framebuffer.\n");
        return false;
    }

    if(displayImage == 0) {
        return false;
    }

    framebufferPitch = screen_width * 4;
    return true;
}

unsigned char* WindowBackend_Software_GetFramebuffer(size_t *pitch)
{
	*pitch = framebufferPitch;

	return (unsigned char*)framebufferPixels;
}

void WindowBackend_Software_DestroyWindow(void) {
    if(useShmP) {
        XShmDetach(xDisplay, &displaySegInfo);
        XShmDetach(xDisplay, &framebufferSegInfo);

        XDestroyImage(displayImage);
        XDestroyImage(framebufferImage);

        shmdt(displaySegInfo.shmaddr);
        shmctl(displaySegInfo.shmid, IPC_RMID, 0);

        shmdt(framebufferSegInfo.shmaddr);
        shmctl(framebufferSegInfo.shmid, IPC_RMID, 0);
    }

    XDestroyWindow(xDisplay, window);
}

void WindowBackend_Software_HandleWindowResize(size_t width, size_t height) {
	(void)width;
	(void)height;

    Backend_PrintError("Bro I can't resize, I'm too lazy.");
}

void WindowBackend_Software_Display(void) {
    if(useShmP) {
        XShmPutImage(xDisplay, window, gc, displayImage, 0, 0, 0, 0, screenWidth, screenHeight, 0);

        // Swap images
        XImage* tempI = displayImage;
        displayImage = framebufferImage;
        framebufferImage = tempI;

        // Swap pixel pointers
        char* tempP = displayPixels;
        displayPixels = framebufferPixels;
        framebufferPixels = tempP;
    } else {
        XPutImage(xDisplay, window, gc, displayImage, 0, 0, 0, 0, screenWidth, screenHeight);
    }
}
